#include "../lib/mysql.hpp"

marcelb::mysql::MySQL::MySQL(const string _path, const string _username, const string _password, const string _db, const uint32_t _available):
#ifdef MYSQL_USE_ASYNCO 
   p_loop(periodic( [&] () {
      cout << "U asynco" << endl;
      try {
         auto start = rtime_ms();
         _tloop(0, connection_pool.size());
         cout << "loop--------------------------- nema error, trajalo: " << rtime_ms() - start << endl;
      } catch (...) {
         // cout << "Bude neki error u loopu" << endl;
         if(on_error) {
            on_error("Bude neki error u loopu");
         }
      }
   }, MYSQL_PERIODIC_INTERNAL_TIME)),
#else 
   tloop_future (async(launch::async, [&](){
      while (run_tloop) {
         cout << "U STD async" << endl;
         usleep(MYSQL_PERIODIC_INTERNAL_TIME*1000);
         _tloop(0, connection_pool.size());
      }
      return;
   })),
#endif
   path(_path),
   username(_username),
   password(_password),
   database(_db),
   pool_size(_available > 0 ? _available : 1) {
   
   drv = get_mysql_driver_instance();
   connect_pool();

   // set on initialization to avoid the error
   last_loop_time = time(nullptr);
}


Connection* marcelb::mysql::MySQL::create_connection() {
   uint32_t trys = 0;
   bool status = true;
   Connection* new_con = NULL;
   while (connect_trys == unlimited ? status : (trys <= connect_trys && status)) {
      try {
         Connection* con_can = drv->connect(path, username, password);
         con_can->setSchema(database);
         status = !con_can->isValid();
         if (!status) {
            new_con = con_can;
         }
         else if (!con_can->isClosed()) {
            disconnect_connection(con_can);
         }
      }
      catch (const SQLException &error) {
         if (on_error) {
            on_error(error.what() + string(", SQL state: ") + error.getSQLState() + string(", Error code: ") + to_string(error.getErrorCode()));
         }
         usleep(reconnectSleep);
         connect_trys == unlimited ? trys : trys++;
      }         
   }

   return new_con;
}

void marcelb::mysql::MySQL::connect_pool() {
   lock_guard<mutex> lock(io);
   for (uint32_t i=0; i<pool_size; i++) {
      Connection* connection = create_connection();
      connection_pool.push(connection);
   }
}

void marcelb::mysql::MySQL::disconnect_pool() {
   lock_guard<mutex> lock(io);
   for (uint32_t i=0; i<connection_pool.size(); i++) {
      Connection* connection = connection_pool.front();
      connection_pool.pop();
      disconnect_connection(connection) ;
   }
}

bool marcelb::mysql::MySQL::disconnect_connection(Connection* connection) {
   bool status = !connection->isClosed();
   if (status) {
      try {
         connection->close();
         status = !connection->isClosed();
      } 
      catch (const SQLException &error) {
         if (on_error) {
            on_error(error.what() + string(", SQL state: ") + error.getSQLState() + string(", Error code: ") + to_string(error.getErrorCode()));
         }
         status = true;
      }
   }

   else {
      status = false; // već je zatvorena
   }

   delete connection;
   return status;
}

void marcelb::mysql::MySQL::_tloop(uint32_t b, uint32_t e) {
   if (!run_tloop) {
      return;
   }
   for (size_t i=b; i<connection_pool.size() && i<e; i++) {
      try {
         Connection *conn = nullptr;
         {
            lock_guard<mutex> lock(io);
            conn = connection_pool.front();
            connection_pool.pop();
         }
         if (conn->isValid()) {
            cout << "Validno----" << endl;
            connection_pool.push(conn);
            condition.notify_one();
         } else {
            cout << "Nije validno----" << endl;
            if (!conn->isClosed()){
               cout << "Zatvori----" << endl;

                  conn->close();
            }
            Connection *n_conn = create_connection();
            {
               cout << "Otvori----" << endl;
               lock_guard<mutex> lock(io);
               connection_pool.push(n_conn);
               condition.notify_one();
            }

         }
      } catch (const SQLException &error) {
         if (on_error) {
            on_error(error.what() + string(", SQL state: ") + error.getSQLState() + string(", Error code: ") + to_string(error.getErrorCode()));
         }
      }
   }

   last_loop_time = time(nullptr);
}

Connection* marcelb::mysql::MySQL::occupy_connection() {
   if (last_loop_time + (MYSQL_PERIODIC_INTERNAL_TIME*3/1000) < time(nullptr)) {
      if (on_error) {
         on_error("The time loop is not executing properly");
      }
   }
   unique_lock<mutex> lock(io);
   while (connection_pool.empty()) {
      condition.wait(lock);
   }
   Connection *connection = connection_pool.front();
   connection_pool.pop();
   return connection;
}

void marcelb::mysql::MySQL::release_connection(Connection* connection) {
   lock_guard<std::mutex> lock(io);
   connection_pool.push(connection);
   condition.notify_one();
}

marcelb::mysql::MySQL::~MySQL() {
   run_tloop = false;
#ifndef MYSQL_USE_ASYNCO
      tloop_future.get();
#endif
   disconnect_pool();
}